# hiton.py
# Mathew Titus, 2021
# 
# 
# 
################################################################

import sys
import numpy as np
from typing import List
from hiton_ezk.ezk import estimate_wasserstein
from hiton_ezk.helper import powerset, c_subselect, sigmoid
from hiton_ezk.data_structures import CpnVar, CpnSub

choleskySetting = True # default: True
symPosSetting = False # default: True


def sample_function(params: dict):
    """
    An example of a simulation function format that is
    compatible with the sample generation needed for EZK
    dependence measure estimation.

    The params dictionary holds all the information needed
    to initialize and simulate the system. A perturbing random
    variable is needed for each variable in `state` to
    perform sampling; this is stored in `distributions`.

    It is recommended that the duration of simulation be one 
    of the parameters for tuning the time between variable 
    perturbation and focal variable response.
    """
    print("Calling sampler with state\n{}".format(params['system']))
    if params['perturbation']:
        print("Calling sample_function with perturbation; variables \
            are sampled by random variables of the form: {}".format([a.__name__ for a in params['distributions']]))
    if 'duration' in params:
        print("Executing simulation for {} time steps.".format(params['duration']))
    # presumably state should change; sample_function just passes state back into params unchanged
    return params | {'state': params['state']}


def l2_agent_distance(agent1: CpnSub, agent2: CpnSub):
    sq_dist = 0.0
    # sum squared differences
    for v in range(len(agent1)):
        sq_dist += (agent1[v].dist(agent1[v].perturbation, agent2[v].perturbation))**2 
    # return the L2 norm of the two vectors w.r.t. the different components metrics
    return np.sqrt(sq_dist)

# NB / TODO:
# In the CpnSub definitions, the entries are treated as
# random variables which have a perturb method that adds
# another, differently distributed, random variable to the
# CpnVar value. In the below we really only need the 
# perturbation distribution, and this is what should be 
# included in the `sampling_params` dictionary.
def define_defaults():
    default_sampling_params = {
        'state': [
            [0, 0], 
            [0, 1], 
            [1, 0], 
            [1, 1]
        ],
        'simulator': sample_function,
        'distributions': [
            [
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                },
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                }
            ],
            [
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                },
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                }
            ],
            [
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                },
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                }
            ],
            [
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                },
                {
                    'variable': np.random.normal,
                    'params': {'loc': 0, 'scale': 0.5},
                    'epsilon': 0.5
                }
            ]
        ],
        'topologies': [
            [
                {'R1': np.nan},
                {'T1': 1}
            ],
            [
                {'R1': np.nan},
                {'T1': 1}
            ],
            [
                {'R1': np.nan},
                {'T1': 1}
            ],
            [
                {'R1': np.nan},
                {'T1': 1}
            ]
        ],
        'perturbation': False,
        'target_variable': [0,0], # [agent index, variable index]
        'duration': 1
    }
    # create default values to be overwritten with user inputs
    default_params = {
        'number_per_inclusion': 1,
        'N': 10,
        'N1': 50,
        'N2': 50,
        'minimum_source_agent_distance': 8/np.sqrt(50 + 50),
        'causal_threshold': 1.0,
        'c_transform': sigmoid,
        'agent_distance': l2_agent_distance,
        'sampling_params': default_sampling_params
    }
    return default_params


def eliminate(x, y, TPC, system, params):
    """
    Use the EZK dependence measure to determine whether
    variable x is independent of y for some subset of TPC.
    This is the elimination step of the HITON interleaving
    algorithm.

    `x` is the single variable that we are interested in learning the causal influences of. 
    It is passed in as a list [x0, x1] with `x0` the member index and `x1` the variable index.

    `y` is an integer, indexing the source member potentially affecting variable `x`.

    `system` comes in as a list of agent states, where the 
    agent state is described by an ordered list of values.
    `cpnsub_system` is a list of CpnSub objects with values
    copied in from `system`, allowing us to make perturb and 
    dist calls.
    """
    assert((x not in TPC) & (y not in TPC)), "Error: source or target agents (x_i = {}, x_j = {}) found in TPC = {}.".format(x,y,TPC)
    simulator = params['sampling_params']['simulator']
    # TODO: here we set the target variable based on input parameters... this should either not change or not be part of the params dictionary.
    params['sampling_params']['target_variable'][0] = x
    # convert the lists of values in `system` to a CpnSub object
    cpnsub_system = []
    for agent_ind in range(len(system)):
        cpn_sub = CpnSub([])
        for var_ind in range(len(system[agent_ind])):
            cpn_var = CpnVar(np.random.uniform, {'low': 0.0, 'high': 1.0}, params['sampling_params']['topologies'][agent_ind][var_ind])
            cpn_var.set_val(system[agent_ind][var_ind])
            cpn_sub.append(cpn_var)
        cpnsub_system.append(cpn_sub)
    # generate samples of entire unconditioned system
    samples = []
    sample_pairs = []
    resamples_required: int = 0
    while len(sample_pairs) < params['N']:
        # copy the system vars
        new_sam = [cpn_sub.copy() for cpn_sub in cpnsub_system]
        # perturb each system var
        for sub_ind in range(len(new_sam)):
            cpn_sub = new_sam[sub_ind]
            for mem_ind in range(len(cpn_sub)): 
                member = cpn_sub[mem_ind]
                member.perturb(
                    params['sampling_params']['distributions'][sub_ind][mem_ind]['variable'], 
                    params['sampling_params']['distributions'][sub_ind][mem_ind]['epsilon'])
        samples.append(new_sam)
        # record which older samples new_sam can be paired with
        if len(samples) == 1:
            continue
        else:
            for old_sam_ind in range(len(samples)-1):
                old_sam_y = samples[old_sam_ind][y]
                new_sam_y = new_sam[y]
                d_y1_y2 = params['agent_distance'](new_sam_y, old_sam_y)
                if d_y1_y2 >= params['minimum_source_agent_distance']:
                    sample_pairs.append((old_sam_ind, len(samples)-1))
                else:
                    resamples_required += 1
                    if resamples_required % 100 == 0: print("{} resamples performed for x={}, y={}...".format(resamples_required,x,y))
    # print("Required {} resamplings to get {} different sample pairs.".format(resamples_required, len(sample_pairs)))
    # list all mediating subsets of TPC
    tpc_subsets = powerset(TPC)
    # test I(x, y | Z) for each subset Z
    for sub in tpc_subsets:
        # for each sample pair, generate conditions K, sample from mu_i distributions, and calculate c_xy^K
        c_xyK = []
        for pair in sample_pairs:
            # samples to calculate W_d^{(k,l)} with
            sample_kay = samples[pair[0]]
            sample_ell = samples[pair[1]]
            # collect variable conditions
            sub_conds_kay = []
            sub_conds_ell = []
            # sample_x is a list of CpnSub's so each needs to be conditioned in seq
            conditioned_cpnsub_kay = []
            conditioned_cpnsub_ell = []
            # calculate distance between j^th agents for later
            d_y1_y2 = params['agent_distance'](sample_kay[y], sample_ell[y])
            for agent_ind in range(len(system)):
                if agent_ind in sub:
                    sub_conds_kay.append(sample_kay[agent_ind].values_to_list())
                    sub_conds_ell.append(sample_kay[agent_ind].values_to_list())
                elif agent_ind == y:
                    sub_conds_kay.append(sample_kay[agent_ind].values_to_list())
                    sub_conds_ell.append(sample_ell[agent_ind].values_to_list())
                else:
                    sub_conds_kay.append([np.nan for _ in range(len(sample_kay[agent_ind]))])
                    sub_conds_ell.append([np.nan for _ in range(len(sample_kay[agent_ind]))])
                # add a copy of agent to conditioned samples
                conditioned_cpnsub_kay.append(sample_kay[agent_ind].copy())
                conditioned_cpnsub_ell.append(sample_ell[agent_ind].copy())
                # perform conditioning
                conditioned_cpnsub_kay[agent_ind].condition(sub_conds_kay[agent_ind])
                conditioned_cpnsub_ell[agent_ind].condition(sub_conds_ell[agent_ind])
            # initialize samples of variable x
            x_draws = np.zeros(params['N1'] + params['N2'])
            # 
            focal_agent = params['sampling_params']['target_variable'][0]
            focal_var = params['sampling_params']['target_variable'][1]
            # draw from \mu_{x}(z^k_{K u y)
            # conditioned_sample = sample_kay.condition(sub_conds_kay)
            for k1 in range(params['N1']):
                for mem_ind in range(len(conditioned_cpnsub_kay)):
                    for var_ind in range(len(conditioned_cpnsub_kay[mem_ind])):
                        # perform perturbation of conditioned system
                        conditioned_cpnsub_kay[mem_ind][var_ind].perturb( 
                            params['sampling_params']['distributions'][mem_ind][var_ind]['variable'], 
                            params['sampling_params']['distributions'][mem_ind][var_ind]['epsilon'] 
                        )
                # generate new simulation results from the new sample 
                new_mu = simulator(params['sampling_params'] | {
                    'state': [cpn_sub.perts_to_list() for cpn_sub in conditioned_cpnsub_kay], 
                    'perturbation': True
                })
                x_draws[k1] = new_mu['state'][focal_agent][focal_var]
            # draw from \mu_{x}(z^ell_{K u y)
            # conditioned_sample = sample_ell.condition(sub_conds_ell)
            for k2 in range(params['N2']):
                for mem_ind in range(len(conditioned_cpnsub_ell)):
                    for var_ind in range(len(conditioned_cpnsub_ell[mem_ind])):
                        # perform perturbation of conditioned system
                        conditioned_cpnsub_ell[mem_ind][var_ind].perturb( 
                            params['sampling_params']['distributions'][mem_ind][var_ind]['variable'], 
                            params['sampling_params']['distributions'][mem_ind][var_ind]['epsilon'] 
                        )
                # generate new simulation results from the new sample 
                new_mu = simulator(params['sampling_params'] | {
                    'state': [cpn_sub.perts_to_list() for cpn_sub in conditioned_cpnsub_ell], 
                    'perturbation': True
                })
                x_draws[params['N1'] + k2] = new_mu['state'][focal_agent][focal_var]
            # calculate dependence measure
            W_kl, _ = estimate_wasserstein(x_draws, params['N1'], params['N2'], cpnsub_system[focal_agent][focal_var])
            c_xyK.append(W_kl / d_y1_y2)
        # # adjust, removing outliers [OPTIONAL]
        # c_xyK = c_subselect(c_xyK)
        if sub == []: c = c_xyK
        # determine whether I(x, y | Z)
        if np.max(c_xyK) < params['causal_threshold']:
            return True, sub, c_xyK
        else:
            continue
    # if I(X,Y|Z) for no subset Z, return False (do not eliminate from TPC)
    return False, [], c


def hiton(focal_index: int, system: List[List], params: dict):
    """
    Apply the HITON interleaving algorithm of Aliferis, Statnikov, et al. (2010)
    to discover the local causal neighborhood of the `focal_index`th agent.
    The function also returns the results of each I(X, Y | Z) calculation;
    blocking_set[y] is a dictionary of { 'Z': c^Z_{x,y} } pairs.
    """
    assert (focal_index < len(system)), "Focal index ({}) must be less than system size ({}).".format(focal, len(system))
    print("Building LCN for agent {}.".format(focal_index))
    # prepare parameter data
    params = define_defaults() | params
    # initializing...
    nagents = len(system)
    blocking_set = {}
    for k in range(nagents): blocking_set[k] = {}
    ## initialize TPC and OPEN, the list of candidate variables
    TPC = []
    OPEN = np.setdiff1d(np.arange(nagents), [focal_index])
    # calculated dependence of focal agent on each other agent with Z = []
    unconditioned_dependence = np.zeros(len(OPEN))
    keep_agent = np.zeros(len(OPEN))
    for j in range(len(OPEN)):
        jay = OPEN[j]
        remove_agent, _, c_ijK = eliminate(focal_index, jay, [], system, params)
        unconditioned_dependence[j] = np.max(c_ijK)
        blocking_set[jay]['null'] = c_ijK
        keep_agent[j] = not remove_agent
    # order OPEN by influence (descending)
    open_order = np.flip(np.argsort(unconditioned_dependence))
    OPEN = OPEN[open_order]
    keep_agent = np.bool_(keep_agent[open_order])
    # discard agents which are not causes of the focal agent
    OPEN = list(OPEN[keep_agent])
    # perform interleaving strategy
    while len(OPEN) > 0:
        print("OPEN: {}\nTPC: {}".format(OPEN, TPC))
        # INCLUSION STEP: transfer most likely agent(s) from OPEN to TPC
        for ns in range(params['number_per_inclusion']): # (eliminate only every number_per_inclusion additions)
            if len(OPEN) > 0:
                TPC.append(int(OPEN[0]))
                agent_to_tpc = OPEN.pop(0) # TODO: record this in a report that can be printed at the end of execution
        # ELIMINATION STEP: test for independence 
        for agent in TPC:
            remove_agent, Z, c_ijK = eliminate(focal_index, int(agent), np.setdiff1d(TPC, agent), system, params)
            # record
            Z.sort()
            blocking_set[agent][str(Z)] = c_ijK
            # eliminate
            if remove_agent:
                TPC.remove(agent)
    print("The final set TPC for agent {} is\n{}".format(focal_index, TPC))
    return TPC, blocking_set


def hiton_iterator(system: List[List], params: dict):
    """
    Call HITON-EZK for each member of the system, results are returned 
    in a dictionary with the parameters used in simulation and the 
    associated blocking variable set (subset K of TPC minimizing c_{i,j}^K)

    `params` holds the parameters for executing the EZK measure approximation.
    `sampling_params` holds the information needed for sampling function to execute;
    in the calls to `eliminate` we will pass samples generated by calling 
    sampling_function(sampling_parameters)

    'sampling_params' has X special key-value pairs that will be defined later including 
    'conditions' which contains a list of lists in the same format as `system`
    """
    assert ((sys.version_info.major == 3) & (sys.version_info.minor >= 9)), "System error: Update Python to a version >= 3.9.1"
    params = define_defaults() | params
    network = {'params': params, 'tpc': {}, 'blocking_set': {}}
    nagents = len(system)
    for focus in np.arange(nagents, dtype=int):
        tpc, Z = hiton(focus, system, params)
        network['tpc'][focus] = tpc
        network['blocking_set'][focus] = Z
    return network


def symmetrize(network):
    """
    Symmetrize the network by ensuring X_i in TPC_j iff X_j in TPC_i.
    This is not used for dynamical simulation where the focal variable 
    X_i cannot influence X_j (as it occurs afterward).
    """
    nagents = len(network['tpc'])
    network['pc'] = {}
    # For each agent
    for aye in range(nagents):
        aye_pc = network['tpc'][aye].copy()
        # Check each potential parent/child
        for jay in network['tpc'][aye]:
            if aye == jay: continue
            # To see if the focal agent is also their parent/child
            if aye not in network['tpc'][jay]:
                aye_pc.remove(jay)
        network['pc'][aye] = aye_pc
    return network










# data_structures.py
# Mathew Titus, November 2020
# 
# Data classes holding the variables of the simulated 
# system. 
# 
# CpnVar contains the variable's value, topology
# distribution when randomly sampled, and notion of 
# distance.
# 
# CpnSub collects a number of CpnVar objects into an 
# array to allow list operations, as well as a 
# condition method, converting random variables to 
# constant values (a new CpnSub object is returned).
# 
######################################################

import numpy as np
from typing import List

## class holding the variable to be processed in an independence test

# possible distributions are in np.random module (uniform, normal, exponential...)
# with parameters given in dict 
# uniform : low, high
# constant : low, high (equal)
# normal : loc, scale

# topology is a dict with key R1, S1, or T1
# and value NaN or, for T1, L with L the length of torus
# R1 = (-inf, inf)
# S1 = [-pi, pi)
# T1 = [0, L)
class CpnVar:
    # constructor
    def __init__(self, variable, params: dict, topology: dict):
        self.rv = variable
        self.parameters = params
        self.value = np.nan
        self.topology = list(topology)[0]
        self.length = topology[self.topology]
        assert len(topology) == 1

    # define the variable's notion of distance
    def dist(self, val1: float, val2: float):
        if self.topology == "R1":
            return np.abs(val2 - val1)
        elif self.topology == "S1":
            sval1 = np.mod(val1, 2*np.pi); sval2 = np.mod(val2, 2*np.pi);
            ccw_dist = np.abs(sval2 - sval1)
            if ccw_dist < np.pi:
                return ccw_dist
            else:
                return 2 * np.pi - ccw_dist
        elif self.topology == "T1":
            tval1 = np.mod(val1, self.length); tval2 = np.mod(val2, self.length);
            euc_dist = np.abs(tval2 - tval1)
            if euc_dist < self.length / 2:
                return euc_dist
            else:
                return self.length - euc_dist
        else:
            raise Exception

    # convert a real input to an appropriate value in the variable domain
    def project(self, val):
        if self.topology == "R1":
            return val
        elif self.topology == "S1":
            s1_val = np.mod(val, 2*np.pi)
            if s1_val >= np.pi:
                s1_val = s1_val - 2*np.pi
            return s1_val
        elif self.topology == "T1":
            t1_val = np.mod(val, self.length)
            return t1_val
        else:
            raise Exception

    # instantiate the variable with value *val*, perturbation defaults to this value also
    def set_val(self, val: float):
        proj_val = self.project(val)
        self.value = proj_val
        self.perturbation = proj_val

    # replace variable value with an independent draw from its null distribution
    def resample(self):
        # some text to see if it shows up in the documentation
        self.set_val( self.rv.__call__(**self.parameters) )
    
    # give an example value of this variable
    def test_sample(self):
        sample = self.rv.__call__(**self.parameters)
        return self.project(sample)
        
    # based on noise and variable type, perturb CpnVar around its value
    def perturb(self, distribution, epsilon: float = 0.1):
        # check that variable is set
        if np.isnan(self.value): 
            print("Cannot perturb as no value is set for this CPN variable.")
            return False
        # sVample according to given distribution, centered at current value
        if distribution.__name__ == "uniform":
            a = self.value - epsilon / 2
            b = self.value + epsilon / 2
            self.perturbation = distribution.__call__(low=a, high=b)
        elif distribution.__name__ == 'normal':
            m = self.value
            s = epsilon
            self.perturbation = distribution.__call__(loc = m, scale = s)
        elif distribution.__name__ == 'randint':
            self.perturbation = distribution.__call__(low=self.params['low'], high=self.params['high'])
        else:
            print("Not a supported perturbation type ({}).".format(distribution.__name__))

    # copy CpnVar contents to a new object
    def copy(self):
        new_copy = CpnVar(self.rv, self.parameters, {self.topology: self.length})
        new_copy.value = self.value
        try:
            new_copy.perturbation = self.perturbation
        except:
            pass
        return new_copy


# CpnSub objects will hold sets of CpnVar elements
class CpnSub:
    # initialize with list of CpnVar objects
    def __init__(self, Members):
        self.members = Members

    # let len be called on CpnSub
    def __len__(self):
        return len(self.members)

    # let CpnSub class be indexed directly
    def __getitem__(self, index: int):
        return self.members[index]

    # let CpnSub elements be replaced by indexed calls
    def __setitem__(self, index: int, data: CpnVar):
        self.members[index] = data

    def copy(self):
        cpy_members = [cpn_var.copy() for cpn_var in self.members]
        cpn_copy = CpnSub(cpy_members)
        return cpn_copy

    # another way to handle element reassignment
    def replace(self, index: int, var: CpnVar):
        self.members[index] = var

    # extend the member set
    def append(self, var: CpnVar):
        self.members.append(var)

    # remove a member by index
    def pop(self, ind: int):
        assert ind < len(self.members), "Index {} is out of range (CpnSub has {} elements).".format(ind, len(self.members))
        self.members.pop(ind)

    # replace random variables in the sub with constant values
    # NaN's indicate an unconditioned variable
    def condition(self, conds: List, pert_params: dict):
        conditioned_cpn = self.copy()
        to_replace = np.where(~np.isnan(conds))[0]
        for ind in np.arange(len(conds)):
            if ind in to_replace:
                # create constant variable
                const_var = CpnVar(
                                np.random.uniform,                          # variable object
                                {'low': conds[ind], 'high': conds[ind]},    # parameters
                                {"R1": np.nan}                              # topology
                            )
                # initialize the CpnVar object at given value
                const_var.resample()
                # replace the old variable
                conditioned_cpn[ind] = const_var
            else:
                # other variables get perturbed
                # print("not conditioning var {}".format(ind))
                cv = conditioned_cpn[ind]
                # print("it begins with perturbation {}".format(cv.perturbation))
                cv_params = pert_params[ind]
                cv.perturb(cv_params['variable'], cv_params['epsilon'])
                # print("it ends with perturbation {}".format(cv.perturbation))
                # print("our CpnSub object holds the value {}".format(conditioned_cpn[ind].perturbation))
        return conditioned_cpn

    # TODO: create this method to alleviate some of the code in `eliminate` in hiton.py
    def perturb(self):
        pass

    # send variable states to a list of constants
    def values_to_list(self):
        out_list = []
        for m in self.members:
            out_list.append(m.value)
        return out_list

    # send variable states to a list of constants
    def perts_to_list(self):
        out_list = []
        for m in self.members:
            out_list.append(m.perturbation)
        return out_list




# helper.py
# Mathew Titus, 2021
# 
# 
# 
#################################

from numpy import min, exp, array, zeros, round, mean
from itertools import combinations


def powerset(original_set, subset_size_limit: int = 5):
    """
    Return the set of all subsets of original_set.
    We omit subsets with size greater than subset_size_limit
    to keep the computational complexity fixed, though users
    may want to adjust this depending on the simulated system
    and computing resources.
    """
    s = list(original_set)
    lim = min((subset_size_limit, len(s))) + 1 
    ps = []
    for r in range(lim): ps.extend([list(d) for d in combinations(s,r)])
    ps.reverse()
    return ps


def sigmoid(value):
    """
    """
    value = array(value)
    s = zeros(value.shape)
    # general form of the sigmoidal function
    sigm = 1*exp(4 * (value - 1/2)) / (1 + exp(4 * (value - 1/2)))
    # the y-value of the corrective line at x=0
    y0 = -1*exp(-2) / (1 + exp(-2))
    # the y-value of the corrective line at x=0
    y1 = 1 - exp(2)/(1 + exp(2))
    interval0 = np.where(value < 0.0)
    interval1 = np.where((value >= 0.0) & (value <= 1.0))
    interval2 = np.where(value > 1.0)
    s[interval0] = 0.0
    s[interval1] = sigm[interval1] + (y1-y0)*value[interval1] + y0
    s[interval2] = sigm[interval2] + y1
    return s


def c_subselect(c_K):
    """
    """
    c_K = np.array(list(c_K.values()))
    msrs = c_K.flatten()[c_K.flatten() > 0]
    msrs.sort()
    return mean(msrs[int(round(0.9*len(msrs))):])





# ezk.py
# Mathew Titus, 2021
# 
# Algorithm for estimating the Wasserstein metric
# 
########################################

import numpy as np
from scipy.optimize import linprog
from hiton_ezk.data_structures import CpnVar, CpnSub


# final argument used to define the distance between points
# for our 1-d cases this is used to calculate the bounds on constraint equations
def estimate_wasserstein(samples, n1, n2, example_var: CpnVar):
    # Define transform matrices
    n: int = n1 + n2
    Ag = np.zeros((n+1,n))
    Bg = np.ones((n,n)) # a "pseudoinverse" of Ag
    # TODO: Calculate Ag & Bg for the R1 topology (formulas below only apply to S1 / T1)
    for i in range(n):
        Ag[i,i] = 1
        Ag[i,np.mod(i+1,n)] = -1
        Ag[n,i] = 1
        for j in range(n):
            Bg[i,np.mod(i+j,n)] = (n-j-1)/n
    # construct objective function
    lf_a = (1.0 / n1) * np.ones((1, n1))
    lf_b = -(1.0 / n2) * np.ones((1, n2))
    linfunct = np.hstack((lf_a, lf_b))[0]
    # reorder sample values and coefficients to ascending order
    ordering = np.argsort(samples)
    samples = samples[ordering]
    f = linfunct[ordering]
    g = np.matmul(Bg.transpose(), f)
    dists = np.zeros(n)
    idists = np.zeros(n)
    # convert to a unit cube with hyperplanar constraint
    if example_var.topology == 'S1':
        for i in range(n):
            j = np.mod(i+1, n)
            d = example_var.dist(samples[i],samples[j])
            dists[i] = d
            idists[i] = 1/d
    elif example_var.topology == 'R1':
        for i in range(n-1):
            j = i+1
            d = example_var.dist(samples[i],samples[j])
            dists[i] = d
            idists[i] = 1/d
    else:
        raise Exception("Error: `estimate_wasserstein` method does not \
            support topology {}!".format(example_var.topology))
    # maximize dot product with this new vector subject to unit cube & constraint dot product
    h = np.matmul(np.diag(dists.flatten()), g)
    # solution alpha must satisfy alpha * dists = 0.
    # find projection of objective vector
    h_tilde = h - (np.dot(h,dists)/np.dot(dists,dists)) * dists
    # arrange system axes so the components reach the boundary in component order
    order_by_mag = np.argsort(-np.abs(h_tilde)) # largest to smallest
    h = h[order_by_mag]
    h_tilde = h_tilde[order_by_mag]
    dists = dists[order_by_mag]
    # sign_vec would be the choice solution within the cube, but violates the orthoganality constraint
    sign_vec = np.sign(h_tilde)
    # working heuristic to avoid saturating too many coordinates leaving no solution set
    cum_alpha_dot_dist = np.abs( np.cumsum(sign_vec * dists) )
    cum_ordered_dists = np.sum(dists) - np.cumsum(dists)
    # find out how many entries we can saturate while retaining a solution set
    # saturated_inds = np.where(cum_ordered_dists < cum_ordered_dists[-1]/2)[0]
    # remaining_inds = np.where(cum_ordered_dists >= cum_ordered_dists[-1]/2)[0] # TODO: replace with np.arange call for speed
    saturated_inds = np.where(cum_alpha_dot_dist < cum_ordered_dists)[0]
    remaining_inds = np.where(cum_alpha_dot_dist >= cum_ordered_dists)[0]
    if len(saturated_inds) == n:
        print("Length autofilled indices: {}\nLength remaining indices: {}\
            \nCalculating final 10\% of entries with linprog".format(len(saturated_inds), len(remaining_inds)))
        saturated_inds = np.array( np.arange(int(np.ceil(0.9 * n))) )
        remaining_inds = np.array( np.setdiff1d(np.arange(n), saturated_inds) )
    # collect saturated coordinate values
    mid_soln = np.zeros(h_tilde.shape)
    mid_soln[saturated_inds] = sign_vec[saturated_inds]
    mid_dp_with_dists = np.dot(mid_soln, dists)
    # use package to fill in remaining values
    h2 = h[remaining_inds]
    d2 = dists[remaining_inds]
    h_t2 = h2.flatten() - (np.dot(h2, d2)/np.dot(d2,d2)) * d2.flatten()
    soln2 = linprog(h_t2, A_eq = d2.reshape((1,-1)), b_eq = mid_dp_with_dists,
        bounds=[(-1,1) for _ in range(len(remaining_inds))])
    # combine
    soln = mid_soln.copy()
    soln[remaining_inds] = -soln2.x
    # get optimal value
    value = np.dot(soln, h)
    # occasionally validate the method
    if np.random.rand() < 1/n:
        val_soln = linprog(h, A_eq = dists.reshape((1,-1)), b_eq = 0,
            bounds=[(-1,1) for _ in range(len(h))])
        print("--- Random Quality Control Call ---\nHeuristic method: \t\t\t{}\nFull linear programming solution:\t{}".format( np.round(value, 2), np.round(-val_soln.fun, 2) ))
    # unwind transformation to get solution vector
    beta_soln = dists * soln
    alpha_soln = np.matmul(Bg, beta_soln.reshape((-1,1)))
    return value, alpha_soln

